

;;;###autoload
(progn
  (setq tcl-application "/user/tjackson/work/tcl/wish2wish.tcl")
  (setq tcl-default-application tcl-application)
  (add-to-list 'hippie-expand-try-functions-list 'try-expand-command-in-remote-tcl-process)
  (defadvice inferior-tcl-mode (before inferior-tcl-mode-requires-w2w protect activate)
    "make sure w2w is loaded before inferior-tcl-mode-requires-w2w"
    (require 'w2w)))

(defun w2w-expand-choice-with-message ()
  "Expand the word under the point if possible"
  (interactive)
  ;; only happens when at the end of a word
  (when (looking-at "\\>")
    (message "yes")))

;;;###autoload
(defun try-expand-command-in-remote-tcl-process (&optional arg)
  "Arg is nil upon first call, and non-nil on subsequent calls
This is to be used in a dabbrev or hippie-expand type of usage."
  (interactive)
  (when (and (memq major-mode '(tcl-mode inferior-tcl-mode))
             (get-buffer-process (if (eq major-mode 'inferior-tcl-mode)
                                     (current-buffer)
                                   inferior-tcl-buffer)))
    (if arg
        (w2w-try-replacements)
      (let (is-variable
            (word-to-expand (w2w-get-current-word)))
        (setq w2w-last-replaced-string nil)
        (setq w2w-replacement-choices nil)
        (when (not (equal "" word-to-expand))
          (if (string-match "^\\$" word-to-expand)
              (progn (setq word-to-expand (substring word-to-expand 1))
                     (setq is-variable t)))
          (setq w2w-last-replaced-string word-to-expand)
          (save-excursion
            (w2w-generate-completion-command word-to-expand is-variable))
          (w2w-try-replacements))))))

(defvar w2w-last-replaced-string nil
  "string of last replacement")
(defvar w2w-replacement-choices nil
  "list of choices from which to replace")

(defun w2w-try-replacements ()
  "Try replacing the current word with one of the choices given"
  (when w2w-replacement-choices
    ;; filter out the replacement that exactly matches the string being expanded
    (let ((replacement (pop w2w-replacement-choices)))
      (while (and replacement (string-equal replacement w2w-last-replaced-string))
        (setq replacement (pop w2w-replacement-choices)))

      ;; when we still have a replacement choice
      (when (and replacement
                 (progn (backward-char (length w2w-last-replaced-string))
                        (looking-at w2w-last-replaced-string)))
        ;; replace it
        (replace-match "")
        (insert (setq w2w-last-replaced-string replacement))
        t))))


(defun w2w-generate-completion-command (initial is-variable)
  "Given an initial string, generate the expression to evaluate in the interpreter"
  ;; only when an inferior tcl is running, and when in a Tcl buffer
  (save-window-excursion
    (w2w-tcl-eval-string "w2w::internal::turn_off_prompt")
     (comint-send-string (inferior-tcl-proc) (concat "tool2wish::try_expand " initial (if is-variable " 1" " ")  "\n")) ; send the query
     (accept-process-output (inferior-tcl-proc))   ; wait for some output
    (w2w-tcl-eval-string "w2w::internal::turn_on_prompt")))

(defun w2w-tcl-eval-string (string)
  "Send the string to the inferior Tcl process."
  (let ((proc (inferior-tcl-proc)))
    (tcl-send-string proc string)
    (tcl-send-string proc "\n")))

(defun w2w-get-current-word ()
  (buffer-substring-no-properties (he-dabbrev-beg) (point)))

(defvar w2w-storage-string "")
(defvar w2w-completed-results-regexp "dynamic-completion-results:\\([^\n]*\\):end-dynamic-completion-results"
  "Regexp that matches the completed results")
(defvar w2w-message-from-tool-begin-regexp "w2w-message-to-emacs:")
(defvar w2w-message-from-tool-end-regexp " \\([^\n ]+\\):end-w2w-message-to-emacs")
(defvar w2w-message-from-tool-regexp "\\(.*\\)mode\\([^\n]*\\):end-w2w-message-to-emacs"
  "Regexp that matches the message from the end tool, 2nd match is the 'type' of the message.")

(defun w2w-filter-completed-results (&optional str)
  "look for tab completed tcl commands"
  (save-match-data
    (let (comint-prompt-read-only)      ;override read only
      (when (string-match w2w-completed-results-regexp str)
        ;; do something with the response
        (let ((results (split-string (match-string 1 str))))
          (setq w2w-replacement-choices results))
        (save-excursion
          ;; delete response
          (goto-char (point-max))
          (when (re-search-backward w2w-completed-results-regexp nil t)
            (replace-match "")
            (delete-char 1)))))))

;; the above should work, but for some reason the string passed in
;; appears to have a limit of 1024 chars.  there's probably a good reason
;; the function doing the result printing is: accept-process-output
;; so look at that some time, in the meantime, don't assume the
;; entire output is in the string.  in fact, when the given string is passed to you
;; the rest of the result may not even be in the process buffer, so queue stuff up
(defun w2w-filter-completed-results (&optional str)
  "look for tab completed tcl commands"
  (save-match-data
    (setq w2w-storage-string (concat w2w-storage-string str))
    (when (not (= (length str) 1024))
      (setq str w2w-storage-string)
      (setq w2w-storage-string nil)
  
      (when (string-match w2w-completed-results-regexp str)
        ;; ok, now look in the buffer for the entire string
        (save-excursion
          (when (re-search-backward w2w-completed-results-regexp nil t)
            (looking-at w2w-completed-results-regexp)
            (let ((b (match-beginning 0))
                  (e (match-end 0))
                  
              ;; delete response
              (delete-region b e)))))))))

(defun w2w-handle-complete (str)
  "given a string of completions, do the right thing"
  (setq w2w-replacement-choices (split-string str)))

(defvar w2w-message-to-function-map
  '(("complete" . w2w-handle-complete)
    ("popup" . w2w-new-frame)
    ("pwd" . w2w-set-pwd)
    ("stacktrace" . w2w-show-stack)
    ("examine" . w2w-view-proc)))

(defun w2w-filter-completed-results (&optional str)
  (save-match-data
    (setq w2w-storage-string (concat w2w-storage-string str))
    (when (and (not (= (length str) 0))
               (not (= (length str) 1024)))
      (setq str w2w-storage-string)
      (setq w2w-storage-string nil)

      ;; make sure we're looking at the latest result
      (goto-char (point-max))
      (when (string-match w2w-message-from-tool-end-regexp str)
        ;; now look in buffer for begin/end pair
        (when (re-search-backward w2w-message-from-tool-begin-regexp nil t)
          (let ((pp (match-beginning 0))
                (p (match-end 0))
                type
                str
                e
                ee)
            (re-search-forward w2w-message-from-tool-end-regexp nil t)
            (setq e (match-beginning 0))
            (setq ee (match-end 0))
            (setq type (match-string 1))
            (setq str (buffer-substring-no-properties p e))
            (delete-region pp (+ 1 ee))

            ;; now call the appropriate function for the type
            (let ((handler (cdr (assoc type w2w-message-to-function-map))))
              (if (and handler (functionp handler))
                  (funcall handler str)
                (error (format "Unknown message received from end tool (type: %s) " type))))))))))

(add-hook 'comint-output-filter-functions 'w2w-filter-completed-results)

(defun w2w-view-proc (str)
  "given a string, set up a buffer that's viewing the string, in Tcl mode"
  (save-excursion
    (set-buffer (get-buffer-create "*Tcl-examine*"))
    (erase-buffer)
    (insert str)
    (insert "\n")
  
    ;; can't make a list with newlines for spacing,
    ;; so let wish2wish signal need for newlines, and we'll provide them
    (goto-char (point-min))
    (while (search-forward "---newlines--- " nil t)
      (replace-match "\n\n\n\n\n"))

    (set-buffer-modified-p nil)
    (tcl-mode)

    ;; pop to buffer when idle
    ;; so as not to mess up the comint filter functions
    (run-with-idle-timer 0 nil (lambda () (switch-to-buffer-other-window "*Tcl-examine*")))))

(defvar w2w-pwd-regexp "\\(/[^ \t\n]+\\) \\([^ \t\n]+\\) ?\\([^ \t\n]*\\)?"
  "The regexp matching the pwd message from the end tool.")

(defun w2w-set-pwd (&optional str)
  "look for the attached process' current path"
  (save-match-data
    (if (string-match w2w-pwd-regexp str)
      ;; do something with the response
      (let ((path (match-string 1 str))
            (host (match-string 2 str))
            (cc_root (or (match-string 3 str) ""))
            (our_cc (or (getenv "CLEARCASE_ROOT") "")))
        (when (and (string-match "^/vobs" path)
                   (not (string= cc_root our_cc)))
          (setq path (concat cc_root path)))
        (when (or (and (not (string= "localhost" host))
                       (w2w-path-is-machine-local path))
                  (not (file-exists-p path)))
          (setq path (format "/%s@%s:%s" (getenv "USER") host path))
          (setq path nil))
        (setq w2w-host host)
        (when path (cd path)))
      (error (format "Path sent in doesn't match the regexp: %s" str)))))

(defun w2w-path-is-machine-local (path)
  "Use /etc/fstab or regexp to see if path is local to a machine."
  (save-match-data
    (let ((shorter-path (if (string-match "/view/[^/]*\\(/.*\\)" path)
                            (match-string 1 path)
                          path)))
      (if (and nil                ; not sure how to use /etc/fstab now
               (file-exists-p "/etc/fstab"))
          (with-temp-buffer
            (insert-file-contents-literally "/etc/fstab")
            (re-search-forward (format "^[a-zA-Z0-9/]* +%s\\b" shorter-path) nil t))
        (string-match "^\\(/tmp\\|/scratch\\|localdisk\\)" shorter-path)))))

(defvar w2w-stack-trace-buffer "*Tcl-stack-trace*"
  "Name of buffer to use for displaying the stack trace")

(defun w2w-get-stack-trace-buffer ()
  "Get/create the stack-trace buffer, and set somet things up inside it."
  (let ((buf (get-buffer-create w2w-stack-trace-buffer)))
    (set-buffer buf)
    (toggle-read-only 1)
    (tcl-stack-trace-settings)
    buf))
  
(defun w2w-show-stack (msg)
  (save-excursion
    (set-buffer (w2w-get-stack-trace-buffer))
    (erase-buffer)
    (insert str)
    (insert "\n")
  
    ;; can't make a list with newlines for spacing,
    ;; so let wish2wish signal need for newlines, and we'll provide them
    (goto-char (point-min))

    (set-buffer-modified-p nil)
    (tcl-mode)

    ;; pop to buffer when idle
    ;; so as not to mess up the comint filter functions
    (run-with-idle-timer 0 nil (lambda () (display-buffer (w2w-get-stack-trace-buffer))))))

(defun w2w-new-frame (msg)
  "Create a new frame."
  (make-frame '((name . "new inferior-tcl-frame"))))

(defun w2w-next-interpreter ()
  "cycle to the next interpreter candidate"
  (interactive)
  (w2w-tcl-eval-string "w2w::next_interpreter"))

(defun w2w-inferior-tcl-mode-hook-function ()
  "stuff to run when inferior-tcl-mode is set up"
  (define-key inferior-tcl-mode-map (kbd "C-M-n") 'w2w-next-interpreter))

(add-hook 'inferior-tcl-mode-hook 'w2w-inferior-tcl-mode-hook-function)


(defvar tcl-proc-marker-alist nil
  "a-list storing tcl procedure/method strings and markers correspondence")

(defun tcl-next-error-function (forward)
  "go to next error in tcl stack trace"
  (interactive "p")
  (let ((money-line-re "\\(procedure\\|method\\|inscope\\) \"\\(.*\\)\" ?\\(script\\|body\\)? line \\([0-9]+\\)")
        (stack-trace-buffer (get-buffer w2w-stack-trace-buffer)))
    (if (not stack-trace-buffer)
        (error "No stack trace"))
    (display-buffer stack-trace-buffer)
    (set-buffer stack-trace-buffer)
    (let ((start-line (line-number-at-pos)))
      (if (or (and (re-search-forward money-line-re nil t forward)
               (not (= start-line (line-number-at-pos))))
              (re-search-forward money-line-re nil t forward))
          (let* ((type (match-string-no-properties 1))
                 (proc (match-string-no-properties 2))
                 (script/body (match-string-no-properties 3))
                 (line (- (string-to-int (match-string-no-properties 4)) 1))
                 (marker (tcl-find-proc-using-tags proc)))
            (when marker
              (with-current-buffer stack-trace-buffer
                (hl-line-highlight))
              (pop-to-buffer (marker-buffer marker))
              (goto-char marker)
              (forward-line line)))
        (error "No more stack.")))))

(eval-after-load "tcl"
  '(progn
       (define-key tcl-mode-map (kbd "<C-up>") 'tcl-up-stack)
       (define-key tcl-mode-map (kbd "<C-down>") 'tcl-down-stack)))

(defun tcl-stack-trace-settings ()
  (require 'hl-line)
  (make-local-variable 'hl-line-face)
  (setq hl-line-face 'tcl-hl-line)
  (copy-face 'hl-line 'tcl-hl-line)
  (set-face-attribute 'tcl-hl-line nil :box t)
  (hl-line-mode 1))

(defun tcl-down-stack ()
  "Move down to the next level of the stack"
  (interactive)
  (tcl-next-error-function 1))

(defun tcl-up-stack ()
  "Move up to the next level of the stack"
  (interactive)
  (tcl-next-error-function -1))

(defvar tcl-hl-line-overlay nil)
(make-variable-buffer-local 'tcl-hl-line-overlay)

(defun tcl-highlight-current-line ()
  (interactive)
  (if tcl-hl-line-overlay
      (delete-overlay tcl-hl-line-overlay))
  (setq tcl-hl-line-overlay (make-overlay 1 1)) ; to be moved
  (overlay-put tcl-hl-line-overlay 'face 'tcl-hl-line)
  (or (pos-visible-in-window-p
       (point) (selected-window))
      (recenter))
  (hl-line-move tcl-hl-line-overlay))

(defun tcl-unhighlight-current-line ()
  (interactive)
  (if tcl-hl-line-overlay
      (delete-overlay tcl-hl-line-overlay)))

(defun tcl-find-proc-using-tags (procname)
  "given a procname, find it using tags, use a cache"
  (let ((match (assoc procname tcl-proc-marker-alist)))
    (if match
        (cdr match)
      (find-tag procname)
      (while (prog2
                 (with-current-buffer (current-buffer)
                   (tcl-highlight-current-line))
                 (not (y-or-n-p (format "Does this match %s ?" procname)))
               (message nil)
               (with-current-buffer (current-buffer)
                 (tcl-unhighlight-current-line)))
        (find-tag procname t))
      (let ((cur (point-marker)))
        (push (cons procname cur) tcl-proc-marker-alist)
        cur))))


(add-to-list 'special-display-buffer-names
             `(,w2w-stack-trace-buffer . ((name . ,w2w-stack-trace-buffer)
                                           ,@default-frame-alist
                                           (left . (- 1))
                                           (top . 0))))


(provide 'w2w)

