#!/bin/sh
#-*-Tcl-*- the next line restarts using wish \
exec /usr/bin/wish  "$0" "$@"


namespace eval tool2wish {
  variable wishname ""
  variable tosendlength 2000
  if {[info exists ::argv0]} {
    set name $::argv0
  } else {
    set name [info nameofexecutable]
  }
  set sourced [expr {[info script] != $name}]
}

array set ::tool2wish::cmd_subst {::print_info Info: ::print_warning Warning: ::print_error Error:}
proc ::tool2wish::grab_output {} {
  foreach cmd [array names ::tool2wish::cmd_subst] {
    if {[info commands $cmd] == {}} {continue}
    
    if {[lsearch [trace info execution $cmd] {enter ::tool2wish::puts_info}] == -1} {
      trace add execution $cmd enter ::tool2wish::puts_info
    }
  }
  if {[lsearch [trace info execution ::puts] {enter ::tool2wish::puts_wrapper}] == -1} {
    trace add execution ::puts enter ::tool2wish::puts_wrapper
  }
}

namespace eval ::t2debug {
  variable debug_server_channel_id ""
}
proc ::t2debug::handle_channel {channel} {
  gets $channel line
  puts "Handling something on channel: $line"
  set code [catch {eval $line} res]
  if {$code} {
    append res " --Error"
  }
  puts $channel $res
  flush $channel
}
proc ::t2debug::debug_server {channel clientaddr clientport} {
  puts "Connection from $clientaddr registered"
  fileevent $channel readable [list ::t2debug::handle_channel $channel]
#  puts $channel [clock format [clock seconds]]
#  close $channel
}
proc ::t2debug::start_server {} {
  variable debug_server_channel_id

  if {$debug_server_channel_id != ""} {
    if {! [catch {eof $debug_server_channel_id}]} {
      puts "Server seems to be running already on id $debug_server_channel_id"
      return
    } else {
      puts "apparently dead server on $debug_server_channel_id"
    }
  }
  puts "bleh"
  if {[catch {set debug_server_channel_id [socket -server ::t2debug::debug_server 9900]} err]} {
    puts "Error while trying to start a t2debug server: $::errorInfo"
  } else {
    puts "Starting a t2debug server on channel $debug_server_channel_id"
    vwait forever
  }
}
proc ::tool2wish::examine_code {glob} {
  set proclist [uplevel 1 [list info commands $glob]]

  set to_send ""
  foreach p $proclist {
    append to_send [get_proc_info_ $p] "\n\n"
  }
  
  send_to_w2w $to_send "examine"
}

proc ::tool2wish::puts_info {args} {
  catch {
    set cmdcall [lindex $args 0]
    set prefix $::tool2wish::cmd_subst([lindex $cmdcall 0])
    set arg1 [lindex $cmdcall 1]
    set arg2 [lindex $cmdcall 2]
    if {$arg2 == {}} { set arg2 global }
    ::tool2wish::puts "{$prefix $arg1 ($arg2)}"
  }
}

proc ::tool2wish::puts_wrapper {args} {
  # have to filter out the puts to files

  # simpler way of doing this is as follows, use it.  fixme
#   set arguments [lrange $cmdstring 1 end]
#   foreach to_ignore {-nonewline stdout stderr} {
#     if {[set idx [lsearch $arguments $to_ignore ]] != -1} {
#       set arguments [lreplace $arguments $idx $idx]
#     }
#   }
#   if {[llength $arguments] <= 1} {
#     # puts was not to a file
#     # so call the original
#     eval ::tkcon_tcl_puts $cmdstring
#   }

#  ::tool2wish::puts [list dude, args is $args]
  catch {
    set stream stdout
    set cmdmsg [lindex $args 0]
    
    set cmdlen [llength $cmdmsg]
    set newargs {}
    switch $cmdlen {
      2 {
        set newargs [lrange $cmdmsg 1 end]; # $message
      }
      3 {
        set unknown [lindex $cmdmsg 1]
        if {[string match -nonewline $unknown]} {
          set newargs [lrange $cmdmsg 1 end]; # -nonewline $message
        } else {
          set stream $unknown
          set newargs [lrange $cmdmsg 2 end]; # stream $message
        }
      }
      4 {
        set stream [lindex $cmdmsg 2]
        set newargs -nonewline
        lappend newargs [lindex $cmdmsg 3]; # -nonewline stream $message
      }
      default {
        error "Unknown argument style passed into puts_wrapper $cmdmsg"
      }
    }
     
    if 0 {
      set msg [lindex $cmdmsg 1]
      if {[llength $cmdmsg] == 3} {
        set stream [lindex $cmdmsg 1]
        set msg [lindex $cmdmsg 2]
      } elseif {[llength $cmdmsg] == 4} {
        # got a -nonewline
        set stream [lindex $cmdmsg 2]
        set msg [lindex $cmdmsg 3]
        set nonewline -nonewline
      }
      puts "wrapper args: {$args}  and parsed:  msg: $msg   stream: $stream   [llength $cmdmsg]"
      puts "wrapper args: $args    parsed to stream: $stream     newargs: $newargs "
      puts "tomatch: [lrange $newargs end end]"
    }
    set topmatch [lrange $newargs end end]
    if {[regexp -- "^\[\{\n\]*w2w" $topmatch]} { 
      return
    }

    if {[lsearch {stdout stderr} $stream] == -1} {
      return
    }

    ::eval ::tool2wish::puts $newargs
  }
}


proc ::tool2wish::puts {args} {
  variable tosendlength
  catch {
    set strtosend $args
    
    if {[string length $args] > $tosendlength} {
      set strtosend {}
      foreach a $args {
        set sub [string range $a 0 $tosendlength]
        append sub " <<trimmed to $tosendlength>>"
        lappend strtosend $sub
      }
    }
    send $::tool2wish::wishname puts $strtosend
    send $::tool2wish::wishname update 
  }
}

#proc ::tool2wish::get_hostname

proc ::tool2wish::send_path {{path {}}} {
  variable wishname
  variable hostname
  
  if {$path == {}} {
    set path [pwd]
  } else {
    set path [file normalize $path]
  }
            
  set ccroot ""
  if {[info exists ::env(CLEARCASE_ROOT)]} {
    set ccroot $::env(CLEARCASE_ROOT)
  }
  
  set from_host [send $wishname [list exec /bin/hostname -s]]
  if {[catch {set from_domain [send $wishname [list exec /bin/hostname -d]]}]} {
    # on aix, this fails
    set from_domain ""
  }
  set from_full ${from_host}.${from_domain}
  
  set host [exec /bin/hostname -s]
  if {[catch {set domain [exec /bin/hostname -d]}]} {
    set domain ""
  }
  set fullhost ${host}.${domain}

  if {$fullhost == $from_full} {
    set fullhost localhost
  }
  
  set hostname $fullhost
  send_to_w2w "$path $fullhost $ccroot" pwd
}

proc ::tool2wish::shiver_me_windows {} {
  # this is used to make all the visible toplevel windows shiver
  # so we know which app we just connected to
  set pause 25
  set delta 5
  set repeat 2
  set wlist [toplist]
  foreach W $wlist {
    set orig_geometry($W) [wm geometry $W]
  }
  while {$repeat > 0} {
    foreach {opx opy} {+ + - + - - + -} {
      foreach W $wlist {
        foreach {x y} [regexp -inline -all -- {\+[0-9]+} $orig_geometry($W)] break
        if {$x == {} || $y == {}} {
          continue
        }
        set newgeo [format "+%d+%d" [expr [list $x $opx $delta]] [expr [list $y $opy $delta]]]
        wm geometry $W $newgeo
        update idletasks
        after $pause
      }
    }
    incr repeat -1
  }
  foreach W $wlist {
    wm geometry $W $orig_geometry($W)
  }
}

proc ::tool2wish::toplist {{W .}} {

  set list {}

  if { [winfo toplevel $W] eq $W } {
    if {! [catch {$W cget -class} res] && ($res eq "Toplevel")} {
      if {([wm transient $W] == {}) && ([wm state $W] eq "normal")} {
        lappend list $W
      }
    }
  }

  foreach w [winfo children $W] {
    set list [concat $list [toplist $w]]
  }

  return $list

}

proc ::tool2wish::parse_error_for_w2w {err info} {
  set didsomething 0
  foreach line [split $info "\n"] {
    if {1 || [regexp {(procedure|method|inscope) "(.*)" ?(script|body)? line ([0-9]+)} $line ignore1 ignore2 procname ignore3 lineno]} {
      append tosend $line "\n"
      continue
      if {! $didsomething} {
        append tosend "\nError: $err\n"
        append tosend "vvvvvvvvvvv error stack vvvvvvvvvvvvvvvvvv\n"
      }
      append tosend "$procname"
      append tosend "  line $lineno\n"
      set didsomething 1
    }
  }
  if {$tosend != ""} { 
    if {[info exists ::iit::quality_feedback::dude_errorstack]} {
      append tosend "\n$::iit::quality_feedback::dude_errorstack"
    }
    append tosend "^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n"
    ::tool2wish::send_to_w2w $tosend stacktrace 
  }
}

proc ::_bgerror err {
  global errorInfo
    
 
  puts "inside [info level [info level]]"
  set l [info level]
  for {incr l -1} {$l > 0} {incr l -1} {
    puts "{---->  proc: '[info level $l]'  namespace: '[uplevel #${l} namespace current]'}"
  }

  #  after idle [list ::queued_error_mail $err $errorInfo]

#  error frog; # so that still get the popup
}


proc ::tool2wish::watch_for_tkerror_rename {args} {
  after 1 [list after idle ::tool2wish::attach_to_error_handling]
}
proc ::tool2wish::error_wrapper {args} {
  after idle [list ::tool2wish::parse_error_for_w2w [lindex $args 0 1] $::errorInfo]
}
proc ::tool2wish::attach_to_error_handling {args} {
  #set wrapwhat bgerror
  if {[package require Tk] <= 8.4} {
    # in 8.4 and before, bgerror is un traceable b/c of bug
    set towrap ::tkerror
    if {[info commands ::tkerror] == ""} {
      proc ::tkerror {args} {
        error
      }
      trace add command ::tkerror {rename delete} ::tool2wish::watch_for_tkerror_rename
    }
  } else {
    set towrap [interp bgerr]
  }
  
  foreach wrapwhat $towrap {
    if {[info commands ${wrapwhat}] == ${wrapwhat} || [auto_load ${wrapwhat}]} {
      set cmd [list enter ::tool2wish::error_wrapper]
      if {[lsearch [trace info execution ${wrapwhat}] $cmd] == -1} {
        ::eval trace add execution ${wrapwhat} $cmd
      }
    }
  }
}
::tool2wish::attach_to_error_handling

proc ::tool2wish::send_to_w2w {msg type} {
  variable wishname
  
  if {$wishname == ""} { 
    # nobody to send info to
    return
  }
  
  send $wishname [list puts -nonewline w2w-message-to-emacs:]
  send $wishname [list puts -nonewline $msg]
  send $wishname [list puts [format " %s:end-w2w-message-to-emacs" $type]]
}

proc ::tool2wish::try_expand {str {is_variable 0}} {
  set res {}
  set str [cmdTruename $str ::]
  if {$is_variable} {
    set res [info vars "$str*"]
  } else {
    set res [info commands "$str*"]
  }
  
  #   set expansionType Procname
  #   if {$is_variable} {
  #     set expansionType Variable
  #   }
  #   foreach t $expansionType {
  #     set expansions [::Widget::Console::Expand$t $str]
  #     ::eval lappend res $expansions
  #   }
  
  send_to_w2w $res "complete"
}

proc ::tool2wish::getProcArglist {p} {
  # given a procedure
  # figure out the argument list
  # as specified when the user defined the procedure
  # e.g. with the defaults
  
  set ns [namespace qualifiers $p]
  set pp [namespace tail $p]
  set args [namespace eval $ns [list info args $pp]]
  set result {}

  if {[catch {
  foreach arg $args {
    set z {}
    if {[namespace eval $ns [list info default $pp $arg z]]} {
      lappend arg $z
    } else {
    }
    lappend result $arg
  }
  }]} {
    set result $args
  }
  return $result
}

proc ::tool2wish::cmdTruename {cmd namespace} {
  #
  # given a command name and a namespace
  # return the fully qualified command name
  # i.e. the name of the command called when
  #    namespace eval $namespace $cmd
  # is evaluated
  if {[regexp {^::} $cmd]} {
    return $cmd
  }
  return [namespace qualifiers ${namespace}::]::$cmd
}

proc ::tool2wish::get_proc_info_ {procname {namespaces ::}} {
  foreach n $namespaces {
    set p [cmdTruename $procname $n]
    if {[info procs $p] != $p} { continue }
  }
  
  set res {}
  
  puts [list namespace eval [list namespace qualifiers $p] [list info body [list namespace tail $p]]]
  lappend res proc $p [getProcArglist $p] [namespace eval [namespace qualifiers $p] [list info body [namespace tail $p]]]
  
  return $res
}

namespace eval ::tool2wish::dbg {
  variable debugger_state_var done
  variable catch_dead_wish2wish 0
}

proc ::tool2wish::dbg::restart_counter {} {
  variable debugger_state_var
  variable catch_dead_wish2wish
  catch {after cancel $catch_dead_wish2wish}
  if {$debugger_state_var != "done"} {
    set catch_dead_wish2wish [after 50000 ::tool2wish::dbg::check_dead_wish2wish]
  } else {
    set catch_dead_wish2wish 0
  }
}

proc ::tool2wish::dbg::check_dead_wish2wish {} {
  set w $::tool2wish::wishname
  if {$w == "" || [catch {send $w {expr 1}}]} {
    set debugger_state_var done
  }
  
  # will do nothing if state is done
  restart_counter
}

proc ::tool2wish::dbg::debug_proc {procname} {
  variable debugger_state_var
  set debugger_state_var stop
  
  set p [uplevel 1 [list ::tool2wish::cmdTruename $procname ::]]
  set traced [trace info execution $p]
  set alreadyTraced 0
  foreach i $traced {
    if {[lindex $i 1] == "::tool2wish::dbg::trace_enter"} {
      set alreadyTraced 1
      break
    }
  }
  if {! $alreadyTraced} {
    trace add execution $p enterstep ::tool2wish::dbg::trace_enter
    trace add execution $p enter ::tool2wish::dbg::trace_enter
  }
  
  restart_counter
  
  return
}

proc ::tool2wish::dbg::undebug_proc {proc} {
  variable debugger_state_var
  set debugger_state_var stop
  
  set p [uplevel 1 {cmdTruename $proc}]
  set traced [trace info execution $p]
  foreach i $traced {
    if {[lindex $i 1] == "::tool2wish::dbg::trace_enter"} {
      trace remove execution $p enterstep ::tool2wish::dbg::trace_enter
      trace remove execution $p enter ::tool2wish::dbg::trace_enter
      break
    }
  }
}

proc stackTrace {{printRoutine puts}} {
  # print stack trace using argument
  set l [info level]
  set app [tk appname]
  for {incr l -1} {$l > 0} {incr l -1} {
    eval $printRoutine [list "$app - $l --->  proc: '[string range [info level $l] 0 1000]'  namespace: '[uplevel #${l} namespace current]'"]
  }
}

proc ::tool2wish::dbg::trace_enter {cmdstring op} {
  variable debugger_state_var
  restart_counter
  
  if {$op == "enterstep" && $debugger_state_var != "continue"} {
    puts "now gonna wait on debugger_state_var after setting state to 'wait'"
    set debugger_state_var wait
    vwait ::tool2wish::dbg::debugger_state_var
  } else {
    puts "$op [info level] $debugger_state_var ---> $cmdstring"
  }
}

proc ::tool2wish::connect {interp {pop {}}} {
  variable wishname
  set wishname $interp
  set cmd ::w2w::connect_to_interp_from_app
  lappend cmd [tk appname]
  send $wishname $cmd
  if {$pop != {}} {
    send $wishname "puts w2w-create-frame"
  }
}

proc ::tool2wish::w2wp {} {
  variable wishname
  if {[catch {send_to_w2w "" "popup"}]} {
    w2w 1
  }
}

proc ::tool2wish::w2we {proc} {
  uplevel 1 [list ::tool2wish::examine_code $proc]
}

proc ::tool2wish::async_eval {cmd} {
  # will be given a command to eval asynchronously
  # send result back by setting a variable
  variable wishname
  
  if {[set code [catch {uplevel 1 $cmd} res]]} {
    set res $::errorInfo
  }
  send $wishname [list set ::w2w::result_code $code]
  send $wishname [list set ::w2w::result $res]
  return
}

proc ::tool2wish::w2w {args} {
  package require Tk
  if {[winfo exists .]} {
    wm withdraw .
  }
  
  set interps [winfo interps]
  set w2s {}
  foreach i $interps {
    if {[string match wish2wish* $i]} {
      lappend w2s $i
    }
  }
  if {[llength $w2s] == 0} {
    error "No wish2wish running found to connect to."
  }
  if {[llength $w2s] == 1} {
    ::tool2wish::connect $w2s $args
    return
  }
  set good_w2ws {}
  foreach w $w2s {
    # just in case one of the wish2wish's is bad, strip it out
    if {! [catch {send $w "puts \"Hey '\[tk appname\]', [tk appname] may connect...\""}]} {
      lappend good_w2ws $w
    }
  }
  set w2s $good_w2ws
  if {[llength $w2s] == 1} {
    connect [lindex $w2s 0] $args; 
    return
  }
  
  # ok, now try filtering out any wish's that are already connected
  set free_w2s {}
  foreach w $w2s {
    if {! [send $w ::w2w::connected_to_tool]} {
      lappend free_w2s $w
    }
  }
  if {[llength $free_w2s] == 1} {connect [lindex $free_w2s 0] $args; return}
  
  # if none available, select from original list
  if {$free_w2s == {}} {
    set free_w2s $w2s
  }
  
  set cmd "tk_dialog .askmedude"
  lappend cmd "wish2wish to connect"
  lappend cmd "Choose a wish2wish to connect to"
  lappend cmd {} -1
  eval lappend cmd $free_w2s
  connect [lindex $free_w2s [eval $cmd]] $args 
}

proc ::tool2wish::respond_to_namechange {command-string code result op} {
  variable wishname
  
  # 'tk appname' wasn't called
  if {! [string match a* [lindex ${command-string} 1]]} { return }
  # result wasn't ok
  if {$code != 0} { return }
  # name was changed, let w2w interpreter know
  # result has the new name
  if {$wishname == ""} { return }
  if {[catch {send $wishname [list set ::w2w::app_interp $result]} err]} {
    puts "wish2wish: Unable to respond to namechange because of error: $err"
  }
}
proc ::tool2wish::register_to_capture_namechange {} {
  set callback [list leave ::tool2wish::respond_to_namechange]
  if {[lsearch [trace info execution ::tk] $callback] == -1} {
    ::eval trace add execution ::tk $callback
  }
}

# go no further if just sourcing the file
if {$::tool2wish::sourced} {
  ::tool2wish::grab_output
  ::tool2wish::register_to_capture_namechange
  
  namespace eval tool2wish {
    namespace export w2w*
  }
  namespace forget w2w*
  namespace import ::tool2wish::w2w*
  
  return
}

namespace eval w2w {}
namespace eval w2w::internal {}
set ::w2w::plain_old_interp 0
set ::w2w::internal::print_prompt 1

set ::w2w::internal::debug [info exists ::env(DEBUG_WISH2WISH)]
proc ::w2w::internal::turn_off_prompt {args} {
  set ::w2w::internal::print_prompt 0
}
proc w2w::internal::turn_on_prompt {args} {
  set ::w2w::internal::print_prompt 1
}

set ::w2w::appname [tk appname wish2wish]

proc ::w2w::connected_to_tool {} {
  # return true if connected to something, other than plain old tcl interp
  variable app_interp
  variable plain_old_interp
  
  # no if just interp
  if {$plain_old_interp} { return 0 }
  
  # no if interp is dead
  if {[catch {send $app_interp {expr 1}}]} {
    return 0
  }
  
  # yes otherwise
  return 1
}

proc ::w2w::connect_to_interp_from_app {interp} {
#  puts "changing b/c app called me $interp"
  return [connect_to_interp $interp]
}

proc ::w2w::connect_to_interp {interp} {
  variable app_interp
  variable plain_old_interp
  
  global n
  
  if {[catch {send $interp "tk appname"} err]} {
    puts stderr "-E- Cannot connect to interpreter!"
    puts stderr "-E- Interpreter might not be a Tcl interpreter:  $interp"
    puts stderr "-E- Actual error was: $err"
    if {$err == "X server insecure (must use xauth-style authorization); command ignored"} {
      puts stderr ""
      puts stderr "You will need to turn off xhost, do a 'xhost -'"
      puts stderr ""
      puts stderr "If you need access from other machines to your display, use 'xauth' instead"
      puts stderr "    see    http://www.acm.uiuc.edu/workshops/cool_unix/xauth.html"
      puts stderr "    the two commands to use are    xauth list        and     xauth add"
    }
    return 0
  }
  
  set n 0
  
  set plain_old_interp [string match "wish2wish*" $interp]
  set app_interp $interp

  if {$interp != $app_interp} {
    error "unable to connect: $::errorInfo"
  }
  puts "w2w -I- Connecting to: $interp"
  
  # FIXME

  dprint "send over $::argv0"
  if {[tk appname] != $interp} {
    send $interp [list source $::argv0]
  }
  
  send $interp ::tool2wish::shiver_me_windows
  #   send $interp $w2w::puts_body
  #   send $interp $w2w::try_expand_body
  dprint "send setting wishname"
  send $interp [list set ::tool2wish::wishname [tk appname]]
  
  # issue command to update path, and update path
  #  send $interp [subst -novariables $w2w::send_path_body]
  dprint "get path"
  catch {send $interp ::tool2wish::send_path}
  
  print_prompt
  #  fileevent stdin readable ::w2w::handle_input
  dprint "get a command"
  fileevent stdin readable ::w2w::get_next_command
  
  return 1
}

proc ::w2w::eval {args} {
  puts [uplevel #0 $args]
}
proc ::w2w::ss {} {
  variable app_interp
  send -async $app_interp ::t2debug::start_server
}
  
proc w2w::next_interpreter {} {
  set interps [interpreter_list]
  if {$interps == {}} {
    set interps [winfo interps]
  }
  ::eval lappend interps $interps
  set idx [lsearch $interps $::w2w::app_interp]
  if {$idx == -1} {
    puts "Warning: couldn't find the current interpreter ($::w2w::app_interp) in the list of interpreters {$interps}"
  }
  while {[incr idx] < [llength $interps]} {
    set interp [lindex $interps $idx]
    # some interps have a send command by a different name
    # just skip them for now
    if {[send $interp {info commands ::send}] != "::send"} { continue }
    dprint "finding next interpreter $interp"
    if {[connect_to_interp $interp]} {
      break
    }
  }
  return ""
}

proc w2w::c {} {
  set ::w2w::app_interp [find_interpreter "Looking for a new app..."]
  dprint "Found app: $::w2w::app_interp"
  connect_to_interp $::w2w::app_interp
  puts ""
  w2w::print_prompt
}

proc w2w::r {} {
  # try reconnecting, if fails, pop up connection dialog
  dprint "trying to reconnect via w2w::r -> $::w2w::app_interp"
  if {[catch {connect_to_interp $::w2w::app_interp}]} {
    ::w2w::c
  }
}

proc w2w::handle_input {next_command} {
  # set next_command [get_next_command]
  
  if {! [regexp {^[ \t\n]*$} $next_command]} {
    set next_command [string trim $next_command]
    # if some controls are being used, simply evaluate them
    # or if redefining w2w commands
    if {[regexp {^(proc )?[:]*w2w::[^ ]*} $next_command evalled]} {
      # puts "W2W: Internal evaluation of: '$evalled'"
      uplevel #0 $next_command
      print_prompt
      return
    }
    
    global n
    
    if {$::w2w::internal::print_prompt} {
      incr n
      w2w::internal::turn_on_prompt
    }
    
    set short_command [string map {"\n" "\n  "} $next_command]
    
    if {[string length $short_command] > 120} {
      set short_command [string range $short_command 0 120]
    }

    if {! $w2w::plain_old_interp && $w2w::internal::print_prompt} {
      dprint "send over the 'running command' string"
      catch {send $w2w::app_interp "puts {\nw2w -I- Running command ($n) sent from wish2wish:\n[quote_meta $short_command]}\n"}
    }
    
    dprint "Next command to eval is ---->${next_command}<-----"
    
    dprint "app interp is? $w2w::app_interp"
    if {[send $w2w::app_interp info commands ::tool2wish::async_eval] == {}} {
      dprint "need to reconnect b/c no async_eval "
      # if a new app of the same name has come up, reconnect
      w2w::r
    }
    dprint "after possible reconnect ($w2w::app_interp)"
    
    if {! [info complete $next_command]} {
      puts "Warning, the command sent is not complete."
      if {[string length $next_command] > 100} {
        puts "[string range $next_command 0 30] ... [string range $next_command end-30 end]"
      } else {
        puts $next_command
      }
    }
    if 1 {
      if {$w2w::plain_old_interp} {
        if {[catch {uplevel #0 $next_command} ::w2w::result]} {
          puts -nonewline "Error:"
        }
        puts $::w2w::result
        print_prompt
        return
      } 
      # while loop so we re-evaluate the given expr in the new interpreter
      # when the interpreter disappears
      dprint "about to send asynchronously the next command to $::w2w::app_interp : [string range $next_command 0 20]"
      while {[catch {send -async $w2w::app_interp [list ::tool2wish::async_eval $next_command]} err]
             && ([string match "no application named*" $err] 
                 ||
                 [string match "receiver never heard of interpreter*" $err])} {
        puts "Error while sending async command to interp: $err"
        set ::w2w::app_interp [find_interpreter "Application $w2w::app_interp gone, please choose another one..."]
      }
      dprint "done sending to $::w2w::app_interp, result is $err"
      
      if {[string match "invalid command name*" $err]} {
        # oops, the target interp has rebooted, need to re-source wish2wish
        dprint "reconnecting b/c of assumed reboot"
        ::w2w::r
      }
      
      # now wait for results
      # why are we doing this asynchronously?
      # because we want wish2wish to respond to debugging stuff
dprint "pre vwait"      
      vwait ::w2w::result
dprint "post vwait"      
      if {$::w2w::result_code} {
        set msg "\nw2w -E- An error occurred!\n${::w2w::result}\n"
        puts stderr $msg
        catch {send $::w2w::app_interp [list puts stderr $msg]}
      } else {
        dprint "print the result {$::w2w::result}"
        if {$::w2w::result != ""} {
          puts $::w2w::result
        }
      }
    }
    
    if 0 {
      if {! [catch {send $w2w::app_interp $next_command} result]} {
        if {$result != ""} {
          puts $result
        }
      } elseif {[string match "no application named*" $result] 
                ||
                [string match "receiver never heard of interpreter*" $result]} {
        set ::w2w::app_interp [find_interpreter "Application $w2w::app_interp gone, please choose another one..."]
      } else {
        set msg "\nw2w -E- An error occurred!\nw2w -E- Error Message:\n$::errorInfo\n"
        puts stderr $msg
        catch {send $w2w::app_interp [list puts stderr $msg]}
      }
    }
  }
  dprint "done with $next_command"
  print_prompt
}

proc w2w::print_prompt {} {
  if {$w2w::internal::print_prompt} {
    puts -nonewline "w2w ($w2w::app_interp)% "
    
    flush stdout
  }
}

proc w2w::interpreter_list {} {
  set interps [winfo interps]
  
  set skip_apps [list \
                     "wish2wish*" \
                     "TdChoose*" \
                     "tclhelp*" \
                     "calibrewb*" \
                     "echorve*" \
                    ]
#  "cgi_app*" \
  
  foreach skip_app $skip_apps {
    while {[set pos [lsearch -glob $interps $skip_app]] != -1} {
      set interps [lreplace $interps $pos $pos]
    }
  }
  
  return $interps
}
proc w2w::find_interpreter {msg} {
  # find all other interpreters
  set interps [interpreter_list]
  
  set this [tk appname]
  
  if {[llength $interps] < 1} {
    puts stderr "w2w -E- No interpreters to connect to, defaulting to a plain old tcl interpreter"
    return $this
  }
  
  set skip_apps [list \
                     "wish2wish*" \
                     "TdChoose*" \
                     "tclhelp*" \
                     "echorve*" \
                    ]
  
  foreach skip_app $skip_apps {
    while {[set pos [lsearch -glob $interps $skip_app]] != -1} {
      set interps [lreplace $interps $pos $pos]
    }
  }
  
  set i 0
  
  set ::w2w::plain_old_interp 0
  catch {destroy .wish2wish}
  set f [toplevel .wish2wish]
  pack [label $f.top -text $msg -relief groove] -fill both -expand yes
  foreach interp [lsort -dictionary $interps] {
    pack [button $f.b$i -text $interp -command "set ::app_interp {$interp}; destroy $f"] -fill both -expand yes
    incr i
  }
  pack [button $f.plain -text "Plain old tcl interpreter..." -command "set ::w2w::plain_old_interp 1; set ::app_interp {$this}; destroy $f"] -fill both -expand yes
  
  center_window_on_screen $f 
  wm deiconify $f
  
  tkwait window $f
  
  return $::app_interp
}

# source /wv/lab_scratch/quartzville/work/tcl/logcmd.tcl
# logcmd::log send

# trace add variable ::w2w::app_interp write ::mywriter 
# proc ::mywriter {args} {
#   upvar [lindex $args 0] myvar
#   puts "was written with value $myvar, here are args: $args"
#   stackTrace
#   if {$myvar == "cgi_app"} {
#     error "not letting you set to cgi_app"
#   }
# }

proc w2w::main {{argv {}}} {
  variable app_interp
  
  if {[winfo exists .]} {
    wm withdraw .
  }
  
  if {[llength $argv] > 0} {
    puts stdout "Using argv: $argv to find interpreter"
    set app_interp [lindex $argv 0]
  } else {
    set app_interp [find_interpreter "Choose an interpreter..."]
  }

  dprint "in w2w::main"
  while {[catch {connect_to_interp $app_interp} result] || !$result} {
    set app_interp [find_interpreter "Error encountered while connecting to interpreter $app_interp...\n$result"]
  }
}

proc w2w::internal::main {} {
  w2w::main
}


proc w2w::center_window_on_screen {toplevel {width ""} {height ""}} {
  update idletasks
  
  if {$width == ""} {
    set width [winfo reqwidth $toplevel]
  }
  if {$height == ""} {
    set height [winfo reqheight $toplevel]
  }
  
  set sw [winfo screenwidth  $toplevel]
  set sh [winfo screenheight $toplevel]
  set vx [winfo vrootx $toplevel]
  set vy [winfo vrooty $toplevel]
  
  set x [expr $sw/2 - $width/2  - $vx]
  set y [expr $sh/2 - $height/2 - $vy]
  
  wm geom $toplevel ${width}x${height}+$x+$y
}


# # don't really want the window, so make it disappear
# wm withdraw .

# proc w2w::connect_to_interpreter {args} {
#   # find all other interpreters
#   set interps [winfo interps]
#   while {[set pos [lsearch -glob $interps ${w2w::appname}*]] != -1} {
#     set interps [lreplace $interps $pos $pos]
#   }

#   # strip out debugger
#   while {[set pos [lsearch -regexp $interps TdChoose|tdb]] != -1} {
#     set interps [lreplace $interps $pos $pos]
#   }

#   # ask to connect to an interpreter if there's more than one to choose from
#   if {[llength $interps] > 1} {
#     set finterps [join $interps "\n"]
#     puts "Please select from the following interpreters: \n$finterps "
#     gets stdin interp
#     while {! (([string is integer -strict $interp] && (($interp < [llength $interps]) || ($interp >= 0)))
#               ||
#               [lsearch -exact $interps $interp] != -1)} {
#       puts "'$interp' and '$interps'"
#       puts "Error, please try again, it must be on of these: $finterps "
#       gets stdin interp
#     }
#   } elseif {[llength $interps] == 1} {
#     # only one interp, connect
#     set interp $interps
#   } else {
#     puts stderr "No interpreters to connect to, please make sure you're running something."
#     exit 1
#   }
#   if {[string is integer -strict $interp]} {
#     set interp [lindex $interps $interp]
#   }

#   puts "Connecting to: $interp"


#   read_eval_print_loop $interp
# }


proc w2w::quote_meta {str} {
  regsub -all {([\"\{\}\[\]\#\$\\])} $str {\\\1} ret
  return $ret
}

proc w2w::dprint {string} {
  if {$::w2w::internal::debug} {
    puts "-W2W:dbg- $string"
  }
}
# proc that reads for next command
variable w2w::cmd ""
proc w2w::get_next_command {} {
  variable cmd
  
  if {[eof stdin]} {
    exit 0
  }
  
  gets stdin line
  append cmd "\n" $line
  
  dprint "Read in line: $line\nwith cmd: $cmd\ncomplete: [info complete $cmd]"
  
  if {[info complete $cmd]} {
    # minor race condition here
    # handle_input lets Tcl process another event
    # which can queue another command up on to the existing command
    # so clear 'cmd' out before we do anything
    set toexe $cmd
    set cmd ""
    if {[catch {w2w::handle_input $toexe} err]} {
      puts "W2W internal error while handling input: $err\nerrorInfo:\n-----------------------\n$::errorInfo\n-----------------------\ntrying to continue..."
    }
  }
  
  dprint "After handling_input, cmd is: $cmd"
  
  #   # repeat until we have a complete command
  #   while {! [info complete $cmd]} {
  #     puts -nonewline "? "
  #     flush stdout
  #     gets stdin line
  #     append cmd "\n" $line
  
  #     dprint "Read in line: $line\nForming: $cmd"
  #   }
  #   dprint "Returning -->${cmd}<-- and [info complete $cmd]"
  #   return $cmd
}

# proc read_eval_print_loop {interp} {
#   # enter read/eval/print loop
#   # hmmm... kinda lispish, eh?
#   global errorInfo

#   while {1} {
#     if {[catch { puts [send $interp [get_next_command]] } err]} {
#       if {[string match "no application named*" $err] 
#           ||
#           [string match "receiver never heard of interpreter*" $err]} {
#         error "Application gone, try again"
#       }
#       puts stdout "$errorInfo"
#       puts stdout "Error: $err"
#     }
#   }
# }

# while {1} {
#   set interp [choose_interpreter]
#   if {! [catch connect_to_interpreter $interp]} {
#     puts "Goodbye...."
#     break
#   }
# }


# while {1} {
#   if {! [catch connect_to_interpreter]} {
#     puts "Exiting read_eval_loop..."
#     break
#   }
# }

w2w::main $argv


## Documentation:
## 
## Just follow the steps below, and you can dynamically attach to any
## running wish process.  And Emacs provides facilities to send
## buffers/strings to processes.  Thus, Emacs can send commands to your
## running tool.
## 
## Just one less reason to ever leave an Emacs session.
## Here's how (tested on both (X)Emacs 20/21):
## 
## 1) copy wish2wish.tcl (attached to this email) to somewhere local
## (make it executable)

## 3) reload the file (this forces Emacs to set the variables)
## 
## 4) Type 'C-c C-t' to start the interpreter (wish2wish.tcl)
## If you have more than one interpreter running, the script
## will ask you for the name of the interpreter to attach to,
## enter it at this point.
## 
## 5) evaluate a region/proc - here are the bindings for (X)Emacs 21
## 
## C-c C-s         switch-to-tcl
## C-c C-x         tcl-eval-region
## C-c C-t         inferior-tcl
## C-c C-f         tcl-load-file
## C-c C-v         tcl-eval-defun
## 
## The results of what you evaluated will appear in the inferior Tcl
## buffer (named "*inferior-tcl*") - as will any error messages.
## However, print statements will not show up.
## 
## 
## 
## 
## ps. If you're still using Emacs 20, you'll need to use a different
## file variable like so:
## 
## And the key bindings have also changed, type "C-h b" to get a list of bindings.
## 
## 
## pps. It's possible you'll get a warning: 
## 
## Error: X server insecure (must use xauth-style authorization); command ignored
## 
## Tk checks to see if your xhost is secure or not.  If it is not
## (i.e. if someone from another machine can connect, Tk disables the
##  'send' command which is what this script uses to connect).  Remove all
## machines from your xhost access list and read the directions here:
## 
## http://www.acm.uiuc.edu/workshops/cool_unix/xauth.html
##
##    xauth list
##    xauth add <paste>
## 
## ppps. Don't type 'exit' to kill the inferior Tcl process.  Just type
## 'C-c C-c' in the process buffer.  'exit' will be sent to the tool and
## exit it.
## 
## pppps. You can also attach from any xterm by just running
## wish2wish.tcl, though I'm not sure what the utility of that is.
## 
## pppps. Let me know if you have any questions. bigfaceworm@gmail.com
## 
